from pathlib import Path


from .error import *
from .judges.kattis import KattisJudge
from .judges.uva import UvaJudge
from .judges.aceptaelreto import AerJudge
from .judges.nonejudge import NoneJudge
from .judges.judge import Judge

def get_judge(judge_name: str, problem: str, path: Path):
    """
    Create a problem from a judge

    Args:
        judge (str): Selected judge
        problem (str): Selected problem
        path (Path): Destination path
    """    

    # Get the judge
    # if the judge is incorrect you get a problem from no judge
    if judge_name == KattisJudge.name:
        return KattisJudge(problem, path)
    elif judge_name == UvaJudge.name:
        return UvaJudge(problem, path)
    elif judge_name == AerJudge.name:
        return AerJudge(problem, path)
    else: 
        return NoneJudge(problem, path)
    
def get_judge_from_file(path: Path) -> Judge:
    """
    Create a problem from a problem folder

    Args:
        path (Path): problem folder
    
    Returns:
        Judge: Desired problem
    """

    # Check if the path exists
    if(not path.is_dir()):
        FileNotFoundError(f"'{path}' folder doesnt' exist")
    
    # Check if it is a valid problem path
    path = Path(path, ".meta")
    if(not path.is_file()):
        raise FileNotFoundError(f"'{path.parent}' folder isn't a problem folder")
    
    # Get data from file
    lines = []
    with(open(path, "r")) as file:
        lines = list(map(lambda x: x[:-1], file.readlines()))

    if(not(len(lines) == 2 and lines[0] != UvaJudge.name or len(lines) == 3 and lines[0] == UvaJudge.name)):
        raise CorruptedMetafileError("Metadata file is corrupted")

    judge = lines[0]
    problem = lines[1]

    # Get judge
    if judge == KattisJudge.name:
        return KattisJudge(problem, path.parent)
    elif judge == AerJudge.name:
        return AerJudge(problem, path.parent)
    elif judge == NoneJudge.name:
        return NoneJudge(problem, path.parent)
    elif judge == UvaJudge.name:
        problem_id = lines[2]
        return UvaJudge(problem, path.parent, problem_id)
    else:
        raise CorruptedMetafileError("Metadata file is corrupted")







import argparse
from pathlib import Path
from requests import exceptions

from ..problem import *
from ..error import InvalidJudgeException, InvalidProblemException
from ..terminal_utils import display_error

def get_display_path(path: Path):

    if(not path.is_absolute()):
        path = Path(Path.cwd().name, path)

    return path

def cli_info():
    
    # Argument parser
    parser = argparse.ArgumentParser(
        prog="cjudge-create",
        description="display information about the problem",
    )

    parser.add_argument(
        "path / judge",
        type=None,
        nargs="?",        
        default=Path("."),
        help="If just one argument is given it must be the problem folder. \nIf two arguments are given this is the selected judge"
    )

    parser.add_argument(
        "problem",
        type=str,
        nargs="?",        
        default=None,
        help="Problem name"
    )
    args: argparse.Namespace = parser.parse_args()
    path = args._get_kwargs()[0][1]
    problem = args.problem

    judge = None
    error_msg = None
    try:    
        if(problem == None):
            path = Path(path)
            judge = get_judge_from_file(path)
        else:
            judge = get_judge(path, problem, None)
            if(judge.name == NoneJudge.name):
                raise InvalidJudgeException(judge)

        judge.display_info()
    except InvalidProblemException as e: 
        error_msg = f"Problem '{e.problem}' isn't a {e.judge} problem"
    except exceptions.HTTPError:
        error_msg = f"Problem '{judge.problem}' isn't a {judge.name} problem"
        if(problem == None):
            error_msg += ". Your metadata file migth be corrupted"
    except (exceptions.ConnectTimeout, exceptions.ConnectionError):
        error_msg = f"Couldn't connect to {judge.name}"
    except InvalidJudgeException as e:
        error_msg = f"'{judge.name}' is an invalid judge"
    except Exception as e:
        error_msg = str(e)

    if(error_msg != None):
        display_error(error_msg)



from pathlib import Path

from .judge import Judge
from ..terminal_utils import *

class NoneJudge(Judge):    

    @property
    def url(self):
        return f"None"

    name = "none"
    fullname = f"{bold}{rgb(Color("#666699"))}None{clear}"

    def __init__(self, problem: str, path: Path):
        self.path = path
        self.problem = problem

    def create_statement(self):
        # No problem statement
        pass
    
    def create_samples(self, force: bool = False, create_sample: bool = True):
        # Create empty input and output
        self.create_samples_empty(force)

    def get_stadistics(self):
        return self.problem, None, None

    def login(self):
        raise NotImplemented
    
    def get_result(self):
        raise NotImplemented

    def submit(self):
        raise NotImplemented


from pathlib import Path
from pypdf import PdfReader
from bs4 import BeautifulSoup
from getpass import getpass
import requests
import json
import time

from ..terminal_utils import *
from ..error import *
from .judge import Judge
from ..error import InvalidProblemException

class UvaJudge(Judge):


    @property
    def url(self):
        return f"https://onlinejudge.org/index.php?option=com_onlinejudge&Itemid=8&page=show_problem&problem={self._problem_id}"

    name = "uva"
    fullname = f"{bold}{rgb(Color("#ff3399"))}UvaJudge{clear}"

    def __init__(self, problem: str, path: Path, problem_id = None):
        self.path = path
        self.problem = problem

        # Problem number must be bigger than 100
        if len(problem) <= 2:
            raise InvalidProblemException(self.name, problem)


        if(problem_id == None):
            # Get problem_id for problem url
            request = requests.get(f"https://uhunt.onlinejudge.org/api/p/num/{self._problem}")
            request.raise_for_status()
            data = json.loads(request.text)
            self._problem_id = data.get("pid")
        else:
            self._problem_id = problem_id

        if(self._problem_id == None):
            raise InvalidProblemException(self.name, problem)

    def create_statement(self): 
        # Get pdf url
        folder = self.problem[:-2]
        pdf_url = f"https://onlinejudge.org/external/{folder}/p{self.problem}.pdf"

        # Download and save pdf directly from uva
        request = requests.get(pdf_url, stream=True)
        request.raise_for_status()

        path = Path(self.path, f"{self.problem}.pdf")
        with open(path, "wb") as file:
            for chunk in request.iter_content(1024):
                file.write(chunk)
    
    def create_samples(self, force: bool = False, create_sample: bool = True):
        # Check if the user want to create empty samples
        if(not create_sample):
            self.create_samples_empty(force)
            return
        
        # Open problem pdf
        reader = PdfReader(Path(self.path, f"{self.problem}.pdf"))

        # Extract text from every page
        text = ""
        for page in reader.pages:
            text += page.extract_text()
            text += "\n"

        # Search for input and output header
        input_index = text.find("Sample Input\n")
        output_index = text[input_index:].find("Sample Output\n")

        # If no sample, we create an empty sample
        if(input_index == -1 or output_index == -1):
            self.create_samples_empty(force)
            return

        output_index += input_index

        # Obtain sample input and output  
        input_text = text[input_index + len("Sample Input"):output_index].strip()
        output_text = text[output_index + len("Sample Output"):].strip()

        # Save sample input and output
        path = Path(self.path, "samples")
        path.mkdir(exist_ok=force)

        with open(Path(path, "1.in"), "w") as file:
            file.write(input_text)

        with open(Path(path, "1.out"), "w") as file:
            file.write(output_text)         

    def create_metadata(self):
        super().create_metadata()
        # We also want to save the problem id
        path = Path(self.path, ".meta")
        with open(path, "a") as file:
            file.write(f"{self._problem_id}\n")

    def get_stadistics(self):
        stadistics_url = self.url.replace("show_problem", "problem_stats").replace("&problem=", "&problemid=")
        request = requests.get(stadistics_url)
        request.raise_for_status()

        parser = BeautifulSoup(request.text, "html.parser")
        parser = parser.find("div", {"id": "col3_content_wrapper"})

        title = parser.find("div", {"class":"componentheading"}).get_text()

        data = parser.find_all("script", limit=4)[-1].get_text()
        first_index = data.find("{")
        second_index = first_index + data[first_index:].find("}") + 1
        parse_data = lambda x: x.replace(")", "").replace("(","").split(" ")
        submission_data = list(map(parse_data, json.loads(data[first_index:second_index].replace("'", '"')).keys()))
        submission_data = {data[0]: int(data[1]) for data in submission_data}

        user_data_tag = parser.find_all("table", limit=2)[-1].find_all("tr", limit=2)[-1].find_all("td")[1:]
        user_data = {"AC": int(user_data_tag[1].get_text()), "WA": int(user_data_tag[0].get_text()) - int(user_data_tag[1].get_text())}

        try:
            submission_data["MLE"] = submission_data.pop("ML")
        except KeyError:
            pass

        try:
            submission_data["TLE"] = submission_data.pop("TL")
        except KeyError:
            pass
        valid_keys = ["AC", "WA", "PE", "MLE", "CE", "TLE", "OT", "RTE"]
        invalid_keys = list(filter(lambda x: not x in valid_keys, submission_data.keys()))
        
        if(len(invalid_keys) > 0):
            submission_data["OT"] = 0
            for key in invalid_keys:
                submission_data["OT"] += submission_data.pop(key)

        return title, user_data, submission_data

    def validate_problem(self):
        url = f"https://uhunt.onlinejudge.org/api/p/num/{self.problem}"
        r = requests.get(url)
        r.raise_for_status()
        
        if(r.text == "{}"):
            raise CorruptedMetafileError

    def login(self):
        self.session = requests.Session()
        post_data = {}

        login_page = self.session.get('https://onlinejudge.org/')
        login_page.raise_for_status()
        soup = BeautifulSoup(login_page.text, "html.parser")

        try:
            form = soup.find('form', attrs={'id': 'mod_loginform'})
            inputs = form.find_all('input')
        except Exception:
            return None
        for i in inputs:
            if i.has_attr('value'):
                post_data[i['name'].encode('ascii','ignore')] = i['value'].encode('ascii','ignore')

        logged = False
        print(f"{bold}Logging in {self.fullname}{clear}")
        print_line()
        for i in range(3):
            user = input("User: ")
            pwd = getpass("Password: ")

            post_data[b"username"] = user.encode("UTF-8")
            post_data[b"passwd"] = pwd.encode("UTF-8")
            post_data[b'remember'] = b'no'
            login_url = form.attrs["action"]

            loader = Loader("Logging in...", "", color=Color("#00FFFFF"))
            loader.start()
            r = self.session.post(login_url, data=post_data)
            r.raise_for_status()

            if(r.text.find("Incorrect username or password. Please try again.") == -1):
                loader.stop()
                logged = True
                break
            
            if(i != 2):
                loader.stop("Incorrect username or password. Please try again")
            else:
                loader.stop("Couldn't log in")


        return logged
    
    def get_result(self):
        entries = None

        loader = Loader("Getting result...", "", color=Color("#00FFFFF"))
        loader.start()
        while entries == None or entries[3].get_text() == "In judge queue":
            time.sleep(0.25)
            r = self.session.get("https://onlinejudge.org/index.php?option=com_onlinejudge&Itemid=9")
            r.raise_for_status()
            parser = BeautifulSoup(r.text, "html.parser")
            entry = parser.find("tr", {"class": "sectiontableentry1"})
            entries = entry.find_all("td")
        
        loader.stop()

        result = entries[3].get_text()
        if result == "Accepted":
            color = rgb(color_dic["AC"])
        else:
            color = rgb(color_dic["WA"])

        time_result = entries[5].get_text()

        print_line()
        print(f"{bold}Result in problem {self.problem}:{clear}")
        print(f"  {bold}{color}{result}{clear} {time_result} secs")

    def submit(self):
        file_path = Path(self.path, "main.cpp")

        if(not file_path.is_file()):
            raise FileNotFoundError
        
        with open(file_path, "r") as file:
            src_code = file.read()

        self.validate_problem()
        if(not self.login()):
            return

        url = "https://onlinejudge.org/index.php?option=com_onlinejudge&Itemid=25&page=save_submission"

        data = {
            'problemid':	'',
            'category':		'',
            'localid':		self.problem,
            'language':		5,
            'code':			src_code
        }

        loader = Loader("Submitting problem...", "", color=Color("#00FFFFF"))
        loader.start()
        r = self.session.post(url, data=data)
        r.raise_for_status()
        loader.stop()

        self.get_result()



from pathlib import Path
import shutil
import json

class Config:
    """
    Class for managing configurations of the app
    """

    config_folder = Path(Path.home(), ".cjudge")
    config_file = Path(config_folder, "config.json")
    template_file = Path(config_folder, "template.cpp")
    package_config_folder = Path(Path(__file__).parent, "config")
    package_config_file = Path(package_config_folder, "config.json")
    package_template_file = Path(package_config_folder, "template.cpp")

    @staticmethod
    def create_config():
        """
        Create config folder and deletes any other previous config
        """
        # Check if config exists and creates it
        if(Config.config_folder.exists):
            shutil.rmtree(Config.config_folder)
        Config.config_folder.mkdir()

        # Move the corresponding files
        shutil.copyfile(Config.package_template_file, Config.template_file)
        shutil.copyfile(Config.package_config_file, Config.config_file)

    @staticmethod
    def repair_config():
        """
        Copies any missing file in the config folder
        """

        # Check if config exists and creates it
        if(not Config.config_folder.exists()):
            Config.config_folder.mkdir()

        package_config_folder = Path(Path(__file__).parent, "config")

        # Check template and config file
        if(not Config.template_file.exists()):
            shutil.copyfile(Config.package_template_file, Config.template_file)

        if(not Config.config_file.exists()):
            shutil.copyfile(Config.package_config_file, Config.config_file)
        

    @staticmethod
    def copy_template(path: Path):
        """
        Copy the code template to a destination path

        Args:
            path (Path): Destination path
        """

        Config.repair_config()
        shutil.copyfile(Config.template_file, path)

    @staticmethod
    def get_config_json():
        config_json = None
        try:
            file = open(Config.config_file, "r")
            config_json = json.load(file)
        except json.decoder.JSONDecodeError:
            shutil.copyfile(Config.package_config_file, Config.config_file)
            config_json = json.load(file)
        finally:
            file.close()
        
        return config_json

    @staticmethod
    def repair_config_parameter(config_json, parameter, new_value = None):
            if new_value == None:
                with open(Config.package_config_file, "r") as file:
                    package_config_json = json.load(file)

                parameter_value = package_config_json.get(parameter)
            else:
                parameter_value = new_value

            config_json[parameter] = parameter_value
            with open(Config.config_file, "w") as file:
                json.dump(config_json, file, indent="\t")

            return parameter_value

    @staticmethod
    def get_compiler():
        """
        Get the C++ compiler
        """

        Config.repair_config()
        config_json = Config.get_config_json()

        compiler = config_json.get("compiler")
        
        if(compiler != None):
            # If the value exist return
            return compiler
        else:
            compiler = Config.repair_config_parameter(config_json, "compiler")
            return compiler
    
    @staticmethod   
    def get_test_output():        
        """
        Get the default test output mode
        """

        Config.repair_config()
        config_json = Config.get_config_json()

        output = config_json.get("test-output")
        choices = ["minimal", "full", "error"]
        if(output in choices):
            return output
        else:
            output = Config.repair_config_parameter(config_json, "test-output")
            return output

    @staticmethod   
    def get_kattis_name():        
        """
        Get the default test output mode
        """

        Config.repair_config()
        config_json = Config.get_config_json()

        output = config_json.get("kattis-username")
        if(output != ""):
            return output
        else:
            output = Config.repair_config_parameter(config_json, "kattis-username", "")
            return None

    @staticmethod   
    def get_kattis_token():        
        """
        Get the default test output mode
        """

        Config.repair_config()
        config_json = Config.get_config_json()

        output = config_json.get("kattis-token")
        if(output != ""):
            return output
        else:
            output = Config.repair_config_parameter(config_json, "kattis-token", "")
            return None


from pathlib import Path
import re
import subprocess
import tempfile
import threading
import time
import queue

from .terminal_utils import *
from .config import Config
from .error import CompilationError

class Tester():
    def __init__(self, path, output, create_files):
        self.cpp_file = Path(path, "main.cpp")
        self.sample_folder = Path(path, "samples")
        self.temporary_path = tempfile.TemporaryDirectory()

        self.output_type = output
        self.create_files = create_files

        self.compile_source()
        self.get_samples()

    def get_samples(self):
        if(not self.sample_folder.exists()):
            raise FileNotFoundError
        
        input_set = set()
        output_set = set()

        for file in self.sample_folder.iterdir():
            name = file.stem
            suffix = file.suffix
            
            if(suffix == ".in"):
                input_set.add(name)
            elif(suffix == ".out"):
                output_set.add(name)

        full_samples = list(input_set.intersection(output_set))
        half_samples = list(input_set.difference(output_set))

        regex = re.compile(r'(\d+)')
        sort_function = lambda l: [int(s) if s.isdigit() else s.lower() for s in re.split(regex, l)]

        full_samples.sort(key=sort_function)
        half_samples.sort(key=sort_function)

        self.samples = []
        for sample in full_samples:
            self.samples.append((True, sample))
        for sample in half_samples:
            self.samples.append((False, sample))

    def compile_source(self):
        loader = Loader("Compiling source...", end_description="", color=Color("#00FFFF"))
        loader.start()
        try:
            self.source = Path(self.temporary_path.name, "main")
            compiler = Config.get_compiler()

            if(not self.cpp_file.exists()):
                raise FileNotFoundError("Couldn't find problem file")

            sub = subprocess.run([compiler, "-o", self.source, self.cpp_file], capture_output=True) 
            if sub.returncode != 0:
                raise CompilationError(sub.stderr.decode())

        finally:
            loader.stop()

    def test_sample(self, sample, full):
        result = None
        answer = None

        input_file = open(Path(self.sample_folder, f"{sample}.in"), "r")
        
        if self.create_files:
            result_file = open(Path(self.sample_folder, f"{sample}.res"), "w+")
        else:
            result_file = tempfile.TemporaryFile("w+")

        if(full):
            output_file = open(Path(self.sample_folder, f"{sample}.out"), "r")

            try:
                sub = subprocess.run(self.source, stdin=input_file, stdout=result_file, timeout=3)
            except subprocess.TimeoutExpired:
                result = "TLE"
            else:
                if(sub.returncode != 0):
                    result = "RTE"

            if(result == None):
                result_file.seek(0)
                is_correct_answer, answer = self.compare_files(result_file, output_file)
                if(is_correct_answer):
                    result = "AC"
                else:
                    result = "WA"
                
            output_file.close()  
        else:
            result = "NI"
            try:
                sub = subprocess.run(self.source, stdin=input_file, stdout=result_file, timeout=3)
            except subprocess.TimeoutExpired:
                result = "TLE"
            else:
                if(sub.returncode != 0):
                    result = "RTE"

            result_file.seek(0)
            answer = [ ('S', result.strip("\n")) for result in result_file.readlines()]

        input_file.close()
        result_file.close()

        return result, answer   

    def compare_files(self, file1, file2):
        strip_endline = lambda x: x.strip("\n")
        file1_lines = list(map(strip_endline, file1.readlines()))
        file2_lines = list(map(strip_endline, file2.readlines()))


        length1 = len(file1_lines) + 1
        length2 = len(file2_lines) + 1

        memo_dis = [[0 for i in range(length1)] for j in range(length2)]
        memo_op = [['E' for i in range(length1)] for j in range(length2)]

        for i in range(length1):
            memo_dis[0][i] = 0
            memo_op[0][i] = 'D'

        for j in range(length2):
            memo_dis[j][0] = 0
            memo_op[j][0] = 'I'

        memo_op[0][0] = 'S'

        for i2 in range(1, length2):
            for i1 in range(1, length1):
                res = memo_dis[i2 - 1][i1]
                op = 'I'

                if(res < memo_dis[i2][i1 - 1]):
                    res = memo_dis[i2][i1 - 1]
                    op = 'D'

                if(file1_lines[i1 - 1] == file2_lines[i2 - 1] and res < memo_dis[i2 - 1][i1 - 1] + 1):
                    res = memo_dis[i2 - 1][i1 - 1] + 1
                    op = 'S'

                memo_dis[i2][i1] = res
                memo_op[i2][i1] = op

        answer = []
        i1 = length1 - 1
        i2 = length2 - 1
        while i1 != 0 or i2 != 0:
            if(memo_op[i2][i1] == 'S'):
                i1 -= 1
                i2 -= 1
                answer.append(('S', file1_lines[i1]))
            elif(memo_op[i2][i1] == 'D'):
                i1 -= 1
                answer.append(('D', file1_lines[i1]))
            else:
                i2 -= 1
                answer.append(('I', file2_lines[i2]))
        answer.reverse()

        is_correct_answer = memo_dis[length2 - 1][length1 - 1] == length1 - 1 and length1 == length2
        return is_correct_answer, answer

    def print_answer(self, answer):
        for op, line in answer:
            color = ""
            char = " "
            if(op == 'D'):
                char = "-"
                color = bold + rgb(Color("#ff7777"))
            elif(op == 'I'):
                char = "+"
                color = bold + rgb(Color("#77ff78"))

            line = line.replace(" ", f"{dim}·{clear}{color}")
            print(f"{color}{char} {line}{clear}")

    def run_tests(self):
        correct_test = 0
        run_test = 0
        fail_test = 0

        # Test full tests
        for full, sample in self.samples:
            loader = Loader(f"Running test {sample}...", end_description="", color=Color("#00FFFF"))
            loader.start()
            result, asnwer = self.test_sample(sample, full)
            loader.stop()
            if(result == "AC"):
                correct_test += 1
                print(f"{bold}Test {sample}: {rgb(color_dic["AC"])}PASSED {check}")
            elif(result == "NI"):
                run_test += 1
                print(f"{bold}Test {sample}: {rgb(color_dic["CE"])}RAN{clear}")
            else:
                fail_test += 1
                print(f"{bold}Test {sample}: {rgb(color_dic["WA"])}FAILED {cross}")

            if(self.output_type == "full" or self.output_type == "error"):
                if(result == "TLE"):
                    print(f"  Your program ran out of time {bold}{rgb(color_dic[result])}{result}{clear}")
                elif(result == "RTE"):
                    print(f"  Your program crashed during execution {bold}{rgb(color_dic[result])}{result}{clear}")
                elif(result == "WA"):
                    self.print_answer(asnwer)
                    print(f"  Your output is incorrect {bold}{rgb(color_dic[result])}{result}{clear}")

            if(self.output_type == "full"):
                if(result == "AC"):
                    self.print_answer(asnwer)
                    print(f"  Your output is correct {bold}{rgb(color_dic[result])}{result}{clear}")
                elif(result == "NI"):
                    self.print_answer(asnwer)
                    print(f"  This is your program output")
            
            print_line()

        summary = []
        if(correct_test > 0):
            summary.append(f"{bold}{rgb(color_dic["AC"])}{correct_test} PASSED{clear}")
        if(fail_test > 0):
            summary.append(f"{bold}{rgb(color_dic["WA"])}{fail_test} FAILED{clear}")
        if(run_test > 0):
            summary.append(f"{bold}{rgb(color_dic["CE"])}{run_test} RAN{clear}")

        if(len(summary) > 0):
            print(f"{bold}Summary: " + f"{bold}, ".join(summary))
        else:
            print(f"{bold}Summary:{clear} No test ran")
        return correct_test, fail_test, run_test

    def run_interactive(self):
        print(f"{bold}Running interactive your interactive program:{clear}")
        print_line()

        def populate_queue(stream, myqueue):
            while True:
                line = stream.readline()
                if(line == None or line == b''):
                    #print("Exit", stream.name)
                    stream.close()
                    break
                else:
                    print(f"{rgb(Color("#9972CC"))}{bold}{line.decode("UTF-8")}{clear}", end="")
                    myqueue.put(line)
                time.sleep(0.001)

        def get_input(stream, thread):
            while True:
                data = input() + "\n"
                stream.write(data.encode("UTF-8"))
                stream.flush()
                time.sleep(0.025)

        process = subprocess.Popen(self.source,stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output_queue = queue.Queue()
        thread1 = Thread(target=populate_queue, daemon=True, args=(process.stdout, output_queue))
        thread2 = Thread(target=populate_queue, daemon=True, args=(process.stderr, output_queue))
        thread3 = Thread(target=get_input, daemon=True, args=(process.stdin, "a"))

        threads = [thread1, thread2, thread3]
        for thread in threads:
            thread.start()

        while(process.poll() == None and threads[0].is_alive):
            time.sleep(0.1)
        
        print_line()
        print(f"{bold}Program exited {process.poll()}{clear}")
        if(thread3.is_alive):
            print("(Press Enter)", end="")

                

        




from pathlib import Path
from requests import exceptions
import argparse

from ..tester import Tester
from ..terminal_utils import *
from ..error import *
from ..config import Config
from ..judges.nonejudge import NoneJudge
from ..problem import get_judge_from_file

def cli_submit():
    # Argument parser
    parser = argparse.ArgumentParser(
        prog="cjudge-submit",
        description="Submit the corresponding problem to the judge",
    )

    parser.add_argument(
        "path",
        type=Path,
        nargs="?",        
        default=Path("."),
        help="The problem folder"
    )

    parser.add_argument(
        "-n", "--notest",
        action="store_true",
        dest="notest",
        default=False,
        help="Don't check your program before sending it"
    )
    args = parser.parse_args()
    path = args.path
    notest = args.notest

    if(not path.exists()):
        display_error("The selected path doesn't exists")
        exit()

    try:
        judge = get_judge_from_file(path)
        if(judge.name == NoneJudge.name):
            raise InvalidJudgeException(judge)

        if(not notest):
            tester = Tester(path, "minimal", False)
            _, wa, _ = tester.run_tests()
            if(wa > 0):
                display_error("Your program didn't pass all test cases")
                exit()
            print("")

        judge.submit()
        
    except CompilationError as e:
        print(e)
        display_error("Couldn't compile your program")
    except FileNotFoundError as e:
        display_error(f"Problem folder '{path}' is not valid")
        display_warning("Check you have a 'samples' folder and a 'main.cpp' file")
    except InvalidProblemException as e: 
        display_error(f"Problem '{e.problem}' isn't a {e.judge} problem")
    except CorruptedMetafileError:
        display_error("Your metadata file is corrupted")
    except exceptions.HTTPError as e:
        error_msg = f"Problem '{judge.problem}' isn't a {judge.name} problem"
        if(problem == None):
            error_msg += ". Your metadata file migth be corrupted"
        display_error(error_msg)
    except (exceptions.ConnectTimeout, exceptions.ConnectionError):
        display_error(f"Couldn't connect to {judge.name}")
    except InvalidJudgeException as e:
        display_error(f"'{judge.name}' is an invalid judge")
    except Exception as e:
        display_error(str(e))


class InvalidJudgeException(Exception):
    """
    Exception raised when a judge is invalid

    Attributes:
        judge (str): selected judge
    """

    def __init__(self, judge: str):
        self.judge = judge
        super().__init__(f"'{judge}' is an invalid judge")

class InvalidProblemException(Exception):
    """
    Exception raised when a problem from a judge doesn't exist

    Attributes:
        problem (str): selected problem
        judge (str): selected judge
    """

    def __init__(self, judge: str, problem: str):
        self.judge = judge
        self.problem = problem
        super().__init__(f"Problem '{problem}' from '{judge}' doesn't exist")

class CorruptedMetafileError(Exception):
    """
    Exception raised when a metafile is not correct
    """

class CompilationError(Exception):
    """
    Exception raised when a compilation error ocurrs
    """


from pathlib import Path
from bs4 import BeautifulSoup
from getpass import getpass
import requests
import json

from ..error import *
from .judge import Judge
from ..terminal_utils import *

class AerJudge(Judge):    
    
    @property
    def url(self):
        return f"https://aceptaelreto.com/problem/statement.php?id={self._problem}"

    name = "aer"
    fullname = f"{bold}{rgb(Color("#99ccff"))}¡Acepta el Reto!{clear}"

    def __init__(self, problem: str, path: Path):
        self.path = path
        self.problem = problem

        # Problem number must be bigger than 100
        if len(problem) <= 2:
            raise InvalidProblemException(self.name, problem)

    def create_statement(self):
        # Get url
        folder = self.problem[:-2].zfill(3)
        pdf_url = f"https://aceptaelreto.com/pub/problems/v{folder}/{self.problem[-2:]}/st/problem.pdf"

        # Download and save pdf directly from Acepta el Reto
        request = requests.get(pdf_url, stream=True)
        request.raise_for_status()

        path = Path(self.path, f"{self.problem}.pdf")
        with open(path, "wb") as file:
            for chunk in request.iter_content(1024):
                file.write(chunk)
    
    def create_samples(self, force: bool = False, create_sample: bool = True):
        # Check if the user want to create empty samples
        if(not create_sample):
            self.create_samples_empty(force)
            return

        # Request problem html
        request = requests.get(self.url, stream=True)
        self.check_if_valid(request)
        parser = BeautifulSoup(request.text, "html.parser")

        # Find input text
        input_tag = parser.find("div", attrs={"id": "sampleIn"}).find("pre")
        input_text = input_tag.get_text()


        # Find output text
        output_tag = parser.find("div", attrs={"id": "sampleOut"}).find("pre")
        output_text = output_tag.get_text()

        # Save sample input and output
        path = Path(self.path, "samples")
        path.mkdir(exist_ok=force)
        
        with open(Path(path, "1.in"), "w") as file:
            file.write(input_text)

        with open(Path(path, "1.out"), "w") as file:
            file.write(output_text)    
    
    def check_if_valid(self, request):
        request.raise_for_status()
        parser = BeautifulSoup(request.text, "html.parser")
        
        button = parser.find("div", {"class": "alert alert-danger"})
        if button != None:
            button_text = button.get_text().strip()
            if button_text.find("ERROR") != -1:
                raise InvalidProblemException(self.name, self.problem)

    def get_stadistics(self):
        stadistics_url = self.url.replace("statement", "statistics")
        request = requests.get(stadistics_url)
        self.check_if_valid(request)

        parser = BeautifulSoup(request.text, "html.parser")
        parser = parser.find("div", {"class": "col-md-10"})

        title = f"{self.problem} - {parser.find("h1").get_text()}"

        user_data_tag = parser.find("table", {"class": "problemGlobalStatistics"}).find_all("tr", limit=2)[-1].find_all("td")[1:]
        user_data = {"AC": int(user_data_tag[1].get_text()), "WA": int(user_data_tag[0].get_text()) - int(user_data_tag[1].get_text())}

        submission_raw_text = parser.find("script").get_text()
        look_index = submission_raw_text.find("];") + 2
        submission_raw_text = submission_raw_text[look_index:]

        first_index = submission_raw_text.find("[")
        second_index = submission_raw_text.find("];") + 1

        submission_raw_data = submission_raw_text[first_index:second_index]
        submission_raw_data = submission_raw_data.replace("[", "").replace("]", "").replace("'", "").split(",")[2:]

        labels_dic = {
            "Accepted": "AC",
            "Presentation Error": "PE",
            "Wrong Answer": "WA",
            "Time limit exceeded": "TLE",
            "Memory limit exceeded": "MLE",
            "Output limit exceeded": "OT",
            "Restricted function": "OT",
            "Run-time error": "RTE",
            "Compilation error": "CE",
            "Internal error": "OT",
        }

        submission_data = {}

        for i in range(0, len(submission_raw_data), 2):
            label = submission_raw_data[i].strip()
            label = labels_dic[label]
            value = int(submission_raw_data[i + 1])

            if(submission_data.get(label) != None and value != 0):
                submission_data[label] += value
            elif (value != 0):
                submission_data[label] = value

        return title, user_data, submission_data

    def login(self):
        self.session = requests.Session()
        post_data = {"commit": "Entrar", "loginForm_currentPage": "/"}

        logged = False
        print(f"{bold}Logging in {self.fullname}{clear}")
        print_line()
        for i in range(3):
            user = input("User: ")
            pwd = getpass("Password: ")

            post_data[b"loginForm_username"] = user.encode("UTF-8")
            post_data[b"loginForm_password"] = pwd.encode("UTF-8")
            login_url = "https://aceptaelreto.com/bin/login.php"

            loader = Loader("Logging in...", "", color=Color("#00FFFFF"))
            loader.start()
            r = self.session.post(login_url, data=post_data)
            r.raise_for_status()

            if(r.text.find("<strong>ERROR: </strong>") == -1):
                loader.stop()
                logged = True
                break
            
            if(i != 2):
                loader.stop("Incorrect username or password. Please try again")
            else:
                loader.stop("Couldn't log in")


        return logged
    
    def get_result(self):
        submission = None

        loader = Loader("Getting result...", "", color=Color("#00FFFFF"))
        loader.start()
        while submission == None or submission["result"] == "IQ":
            time.sleep(0.25)
            r = self.session.get("https://aceptaelreto.com/ws/user/30966/submissions")
            r.raise_for_status()
            data = json.loads(r.text)
            submission = data["submission"][0]


        loader.stop()

        result = submission["result"]
        time_result = submission.get("executionTime")
        memory = submission.get("memoryUsed")
        ranking = submission.get("ranking")

        color = rgb(color_dic.get(result, color_dic["OT"]))
        result = veredict_dict.get(result, result)
        result_str = f"  {bold}{color}{result}{clear}"

        if(time_result != None):
            result_str += f" {time_result} secs"
        
        if(memory != None):
            result_str += f" {memory} KiB"
        
        if(ranking != None):
            result_str += f" {bold}#{ranking}{clear}"

        print_line()
        print(f"{bold}Result in problem {self.problem}:{clear}")
        print(result_str)

    def submit(self):
        file_path = Path(self.path, "main.cpp")

        if(not file_path.is_file()):
            raise FileNotFoundError
        
        with open(file_path, "r") as file:
            src_code = file.read()

        if(not self.login()):
            return

        url = f"https://aceptaelreto.com/bin/submitproblem.php"

        data = {
            "currentPage": (None, "/"),
            "cat": (None, -1),
            "id": (None, self.problem),
            "language": (None, "CPP"),
            "comment": (None, ""),
            "inputFile": ("", ""),
            "sentCode": (None, "inlineSentCode"),
            "immediateCode": (None, src_code),
        }

        loader = Loader("Submitting problem...", "", color=Color("#00FFFFF"))
        loader.start()
        r = self.session.post(url, files=data)
        r.raise_for_status()
        loader.stop()

        if(r.text.find("El problema no existe") != -1):
            raise CorruptedMetafileError

        self.get_result()


from pathlib import Path
import argparse

from ..tester import Tester
from ..terminal_utils import *
from ..error import *
from ..config import Config

def cli_test():
    # Argument parser
    parser = argparse.ArgumentParser(
        prog="cjudge-test",
        description="Run problem test cases from 'samples' folder",
    )

    parser.add_argument(
        "path",
        type=Path,
        nargs="?",        
        default=Path("."),
        help="The problem folder"
    )

    parser.add_argument(
        "-o", "--output",
        type=str,
        metavar="choice",
        dest="output",
        choices=["minimal", "error", "full"],
        default=None,
        help="Information to be displayed. quiet: Only test results. error: Error outputs. full: All the outputs. (Default can be changed on config file)"
    )

    parser.add_argument(
        "-i", "--interactive",
        action="store_true",
        dest="interactive",
        default=False,
        help="Run an interactive test case"
    )

    parser.add_argument(
        "-nf", "--nofile",
        action="store_false",
        dest="create_files",
        default=True,
        help="Doesn't create result files"
    )

    args = parser.parse_args()
    output = args.output
    interactive = args.interactive
    create_files = args.create_files
    path = args.path

    if(output == None):
        output = Config.get_test_output()

    if(not path.exists()):
        display_error("The selected path doesn't exists")
        exit()

    try:
        tester = Tester(path, output, create_files)
        tester.run_tests()

    except CompilationError as e:
        print(e)
        display_error("Couldn't compile your program")
    except FileNotFoundError as e:
        display_error(f"Problem folder '{path}' is not valid")
        display_warning("Check you have a 'samples' folder and a 'main.cpp' file")



import argparse
from pathlib import Path
from requests import exceptions
from shutil import copytree, rmtree

from ..judges.judge import Judge
from ..problem import get_judge
from ..error import *
from ..terminal_utils import *


def cli_create():
    parser = argparse.ArgumentParser(
        prog="cjudge-create",
        description="create a problem from an online judge",
    )

    parser.add_argument(
        "judge",
        metavar="judge",
        nargs="?",
        help="online judge from 'uva', 'aer', 'kattis'. If none leave empty",
    )

    parser.add_argument(
        "problem",
        metavar="problem",
        help="selected problem",
    )

    parser.add_argument(
        "-p", "--path",
        type=Path,
        default=None,
        dest="path",
        help="problem destination location"
    )

    parser.add_argument(
        '-f', 
        '--force', 
        default=False, 
        action='store_true',
        dest="force",
        help="force the creation of the problem even if the folder exists"
    )

    parser.add_argument(
        '--nostatement',  
        default=True, 
        action='store_false',
        dest="statement",
        help="Doesn't create the problem statement"
    )

    parser.add_argument(
        '--nosample',  
        default=True, 
        action='store_false',
        dest="sample",
        help="Create a sample folder with empty sample cases"
    )

    # Get cli arguments
    args = parser.parse_args()
    
    path: Path = args.path
    force: bool = args.force
    sample: bool = args.sample
    statement: bool = args.statement
    problem: str = args.problem
    judge_name: str = args.judge

    # Default path
    if(path == None):
        path = Path(Path(".", args.problem))

    # Start the creation of the problem
    loader = Loader(
        "Creating folder...", 
        f"Problem created sucessfully {check}",
        color=Color("#00ffff")
    )
    loader.start()

    error_msg = None
    remove_dir = False


    try:
        # Get problem
        judge = get_judge(judge_name, problem, path)

        # Check if the judge was correct and warn the user
        if(judge.name == "none" and judge_name != None):
            loader.send_warning("Invalid judge creating an empty problem")

        # Create the problem folder
        path.mkdir(parents=True, exist_ok=force)
        remove_dir = True

        # Download problem statement
        if(statement):
            loader.change_description("Downloading statement...")
            judge.create_statement()

        # Download problem samples
        if(sample):
            loader.change_description("Downloading samples...")
        judge.create_samples(force, sample)

        # Create extra files
        loader.change_description("Creating extra files...")
        judge.create_template()
        judge.create_metadata()

    # Error handling
    except FileExistsError:
        error_msg = "Folder already exists"
    except (exceptions.HTTPError, InvalidProblemException):
        error_msg = f"Problem '{problem}' isn't a {judge_name} problem"
    except (exceptions.ConnectTimeout, exceptions.ConnectionError):
        error_msg = f"Couldn't connect to {judge_name}"
    except Exception as e:
        error_msg = str(e)
    else:    
        remove_dir = False

    # Clean up
    if(remove_dir):
        rmtree(path)
    loader.stop(error_msg)


from abc import ABC, abstractmethod
from pathlib import Path
from requests import exceptions
import shutil 

from ..error import *
from ..terminal_utils import *
from ..config import Config

class Judge(ABC):
    """
        Abstract Class that represent a judge
    """

    @property
    @abstractmethod
    def name(self):
        """Judge short name"""
        raise NotImplementedError

    @property
    @abstractmethod
    def fullname(self):
        """Judge fullname"""
        raise NotImplementedError

    @property
    @abstractmethod
    def url(self):
        """Problem url"""
        raise NotImplementedError  

    @property
    def problem(self):
        return self._problem

    @problem.setter
    def problem(self, problem: str):
        self._problem = problem

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path: Path):
        self._path = path

    @abstractmethod
    def __init__(problem: str, path: Path):
        """
        Create a Judge with an associated problem

        Args:
            problem (str): problem id of the judge
            path (Path): Problem destination path
        """
        raise NotImplementedError

    @abstractmethod
    def create_statement(self):
        """
        Download problem statement in a path
        """
        raise NotImplementedError

    @abstractmethod
    def create_samples(self, force: bool = False, create_sample: bool = True):
        """
        Create problem samples in a desired folder

        Args:
            force(bool): forces to continue even if the directory exists
            create_sample(bool): False in order to create empty samples
        """
        raise NotImplementedError
    
    def create_samples_empty(self, force: bool = False):
        """
        Create empty samples
        Args:
            force(bool): forces to continue even if the directory exists
        """

        path = Path(self.path, "samples")
        path.mkdir(exist_ok=force)

        with open(Path(path, "1.in"), "w") as file:
            pass

    def create_template(self):
        """
        Create a template from config folder.
        If it cant get the template it reconstructs the config folder
        Args:
            path (Path): destination path
        """   

        path = Path(self.path, "main.cpp")
        Config.copy_template(path)
        
    def create_metadata(self):
        """
        Create a metadata file
        Args:
            path (Path): destination path
        """   
        
        path = Path(self.path, ".meta")
        with open(path, "w") as file:
            file.write(self.name + "\n")
            file.write(self.problem + "\n")

    def display_info(self):
        """Display info to stdout"""

        loader = Loader("Getting problem information...", "", color=Color("#00ffff"))
        loader.start()
        try:
            title, user, submission = self.get_stadistics()
        finally:
            loader.stop()

        print(f" {bold}{title}{clear} ".center(113, "┈"))
        print(f"{bold}Judge:{clear}   {self.fullname}")
        print(f"{bold}Problem:{clear} {self.problem}")
        print(f"{bold}URL:{clear}     {self.url}")

        if(user != None and submission != None):
            submission_data = [(veredict, value, color_dic[veredict]) for veredict, value in submission.items()]
            submission_data = list(zip(*submission_data))
            submission_bar = Bar(names=submission_data[0], values=submission_data[1], colors=submission_data[2], title="Submissions:")

            user_data = [(veredict, value, color_dic[veredict]) for veredict, value in user.items()]
            user_data = list(zip(*user_data))
            user_bar = Bar(names=user_data[0], values=user_data[1], colors=user_data[2], title="Users:      ")

            print(f" {bold}Statistics{clear} ".center(113, "┈"))
            user_bar.display(show_legend=False, columns=105)
            print("")
            submission_bar.display(show_legend=True, columns=105)


    @abstractmethod
    def get_stadistics(self):
        """
        Gets problem stadistics and return them
        """
        raise NotImplementedError

    @abstractmethod
    def login(self):
        """
        Returns a logged session to the judge
        """
        raise NotImplementedError

    @abstractmethod
    def get_result(self):
        """
        Get result from a problem submission
        """
        raise NotImplementedError

    @abstractmethod
    def submit(self):
        """
        Submit problem to the judge
        """
        raise NotImplementedError


from threading import Thread
from shutil import get_terminal_size
import time

class Color():
    def __init__(self, *args):
        if(len(args) == 3):
            self.r = args[0]
            self.g = args[1]
            self.b = args[2]
        elif(len(args) == 1 and type(args[0]) == str):

            string = args[0].lower()
            if(string[0] == "#"):
                string = string[1:]
            hex_int = int(string, base=16)
            self.r = hex_int // (256 ** 2)        
            self.g = (hex_int // 256) % 256  
            self.b = hex_int % 256
        elif(len(args) == 1):
            rgb_list = args[0]
            self.r = args[0]
            self.g = args[1]
            self.b = args[2]  
        else:
            raise ValueError("Arguments must be r, g, b integers or one rgb string")
    
    def __repr__(self):
        return f"Color({self.r},{self.g},{self.b})"

def rgb(color: Color) -> str:
    """
    Args:
        r (int): red
        g (int): green
        b (int): blue

    Returns:
        str: Return rgb ansi escape code
    """
    return f"\33[38;2;{color.r};{color.g};{color.b}m"

def rgb_bg(color: Color) -> str:
    """
    Args:
        r (int): red
        g (int): green
        b (int): blue

    Returns:
        str: Return rgb background ansi escape code
    """
    return f"\33[48;2;{color.r};{color.g};{color.b}m"

clear = "\33[0m"
bold = "\33[1m"
italic = "\33[3m"
dim = "\33[2m"

cross = f"{rgb(Color("#E84F67"))}✘{clear}"
check = f"{rgb(Color("#55B369"))}{bold}✓{clear}"

color_dic = {
    'AC':Color('#55B369'),
    'WA':Color('#E84F67'),
    'TLE':Color('#F3B74D'),
    'TL':Color('#F3B74D'),
    'MLE':Color('#75A9D4'),
    'ML':Color('#75A9D4'),
    'CE':Color('#C45A9C'),
    'PE':Color('#FF9966'),
    'RTE':Color('#9972CC'),
    'OT':Color('#000000')
}

veredict_dict = {
    "AC": "Accepted",
    "WA": "Wrong Answer",
    "TLE": "Time limit Exceeded",
    "MLE": "Memory limit Exceeded",
    "CE": "Compile Error",
    "PE": "Presentation Error",
    "RTE": "Run-Time Error",
    "OT": "Other",
    "TL": "Time limit Exceeded",
    "MLE": "Memory limit Exceeded"

}

def print_line():
    print(min(get_terminal_size().columns, 113) * "┈")

def underline(msg: str, color: Color):
    columns = get_terminal_size().columns
    msg_length = len(msg)
    print(f"{rgb_bg(color)}{msg}{" " * (columns - msg_length)}{clear}")

def display_warning(msg: str, sameline: bool = False):
    """Display a warning msg"""
    error_color = Color("#F3B74D")
    c = ""
    if(sameline):
        c = "\r"
    print(f"{c}{bold}{rgb(error_color)}[WARNING]{clear} {msg}")

def display_error(msg: str):
    """Display a warning msg"""
    error_color = Color("#E84F67")
    print(f"{bold}{rgb(error_color)}[ERROR]{clear} {msg}")

class Loader():
    def __init__(self, description: str, end_description: str, sequence: list = ["⡇","⠏", "⠛", "⠹", "⢸", "⣰", "⣤", "⣆"], color = None):
        self.description = description
        self.end_description = end_description
        self.sequence = sequence
        self.done = False
        self.thread = Thread(target=self._animate, daemon=True)
        self.color = color

    def start(self):
        self.done = False
        self.thread.start()

    def stop(self, error = None):
        self.done = True
        print(f"\r{ " " * get_terminal_size((80, 20)).columns }", end="")

        if(error == None):
            if(self.end_description == ""):
                print("\r", end="")
            else:
                print(f"\r{self.end_description}")
        else:
            print(f"\r", end="")
            display_error(error)

    def change_description(self, description: str):
        self.description = description
    
    def send_warning(self, description: str):
        display_warning(description, True)

    def _animate(self):
        index = 0
        cycle_length = len(self.sequence)
        last_description = self.description

        color = rgb(self.color) if self.color != None else ""

        while not self.done:

            # If the description has changed we clear the line before printing
            if(self.description != last_description):
                print(f"\r{ " " * get_terminal_size((80, 20)).columns }", end="", flush = False)

            print(f"\r{color}{self.sequence[index]}{clear} {self.description}", end="", flush=True)
            index = (index + 1) % cycle_length
            last_description = self.description
            time.sleep(0.1)
    
    def __del__(self):
        if(not self.done):
            self.stop() 
        
class Bar:
    def __init__(self, values, names, colors, title = None):
        total_sum = sum(values)
        self.values = [value / total_sum for value in  values]
        self.colors = colors
        self.names = names
        self.title = title
    
    def display(self, show_legend: bool = True, columns = 100):
        # Offset due to the title
        title_offset = len(f"{self.title} ")
        if(self.title != None):
            columns -= title_offset 

        # Convert percentages to number of characters
        ceil_chars_index = []
        lower_chars_index = []
        number_chars = []
        chars_number_sum = 0

        for index, value in enumerate(self.values):
            value *= columns
            rounded_value = round(value)

            chars_number_sum += rounded_value
            number_chars.append(rounded_value)

            if(rounded_value > value):
                ceil_chars_index.append(index)
            elif(rounded_value < value):
                lower_chars_index.append(index)

        if(chars_number_sum < columns):
            number = columns - chars_number_sum
            lower_chars_index.sort(key=lambda x: self.values[x], reverse=True)
            for i in range(number):
                index = lower_chars_index[i]
                number_chars[index] -= 1

        elif(chars_number_sum > columns):
            number = chars_number_sum - columns 
            ceil_chars_index.sort(key=lambda x: self.values[x], reverse=True)
            for i in range(number):
                index = ceil_chars_index[i]
                number_chars[index] -= 1

        # Get bar string
        color = []
        for i, value in enumerate(number_chars):
            color += value * [self.colors[i]]
        bar = ""
        for i in range(0, len(color)):
            char = "█"
            if(i in [0,len(color) - 1]):
                char = "░"
            elif(i in [1,len(color) - 2]):
                char = "▒"
            elif(i in [2,len(color) - 3]):
                char = "▓"
            bar += rgb(color[i]) + char
        bar += clear

        # Show bar
        if(self.title != None):
            print(f"{bold}{self.title} {clear}", end="")    
        print(bar)
        
        # Show legend
        if(show_legend):
            # Calculate center offset and create legend
            legend = ""
            length = 0
            for color, value, name, char in zip(self.colors, self.values, self.names, number_chars):
                percent = str(round(value * 100, 2))
                length += len(name + percent) + 4
                legend += f"{bold}{rgb(color)}{name}:{clear}{percent}%  "
            length -= 2

            # Print title and center offset
            print("\n" + " " * (title_offset + (columns - length) // 2), end="")
            # Print legend
            print(legend)

    


from pathlib import Path
from zipfile import ZipFile 
from bs4 import BeautifulSoup
import requests
import kattispdf
import shutil

from ..terminal_utils import *
from ..error import *
from ..config import Config
from .judge import Judge


class KattisJudge(Judge):
    @property
    def url(self):
        return f"https://open.kattis.com/problems/{self._problem}"

    name = "kattis"
    fullname = f"{bold}{rgb(Color("#ff9933"))}Kattis{clear}"

    def __init__(self, problem: str, path: Path):
        self.path = path
        self.problem = problem

    def create_statement(self):
        path = Path(self.path, f"{self.problem}.pdf")
        kattispdf.generate_pdf(self.problem, path)
    
    def create_samples(self, force: bool = False, create_sample: bool = True):
        
        # Check if the user want to create empty samples
        if(not create_sample):
            self.create_samples_empty(force)
            return

        # Request sample
        samples_url = f"{self.url}/file/statement/samples.zip"
        request = requests.get(samples_url, stream=True)

        # If problem not found create empty samples
        if(request.status_code == 404):
            self.create_samples_empty(force)
            return
        request.raise_for_status()

        # Create folder
        path = Path(self.path, "samples")
        path.mkdir(exist_ok=force)

        # Download samples
        zip_path = Path(path, "samples.zip")
        with open(zip_path, "wb") as file:
            for chunk in request.iter_content(1024):
                file.write(chunk)

        # Extract samples
        with ZipFile(zip_path, 'r') as zip_file:  
            zip_file.extractall(path) 

        # Clean up download
        zip_path.unlink()

        # If no sample, we create an empty sample
        if(len(list(path.iterdir())) == 0):
            with open(Path(path, "1.in"), "w") as file:
                pass
            with open(Path(path, "1.out"), "w") as file:
                pass

        # Change .ans for .out
        for file in path.glob("*.ans"):
            file.rename(file.with_suffix(".out"))

    def get_stadistics(self):
        stadistics_url = self.url + "/statistics"
        request = requests.get(stadistics_url)
        request.raise_for_status()

        parser = BeautifulSoup(request.text, "html.parser")

        title = parser.find("h1").get_text().strip()

        parser = parser.find("div", {"class": "flex flex-wrap w-full gap-3"})
        data = parser.find("div", {"class": "w-full basis-full basis-2/3@md mt-5"}).find("script").get_text()
        first_index = data.find("[")
        second_index = first_index + data[first_index:].find("]") + 1

        submission_data = data[first_index:second_index].strip("][").replace("'", "").split(',')
        submission_data = [ int(value) for value in submission_data ]

        first_index = data[second_index:].find("[") + second_index
        second_index = first_index + data[first_index:].find("]") + 1

        submission_labels = data[first_index:second_index].strip("][").replace("'", "").split(',')

        labels_dic = {
            "Accepted": "AC",
            "Wrong Answer": "WA",
            "Run-Time Error": "RTE",
            "Compile Error": "CE",
            "Time Limit Exceeded": "TLE",
            "Memory Limit Exceeded": "MLE",
            "Other": "OT",
            "Output Limit Exceeded": "OT",
            "Judge Error": "OT",
        }

        submission_labels = [labels_dic[label] for label in submission_labels ]
        submission_data = {label:value for value, label in zip(submission_data, submission_labels)}

        table = parser.find("table", {"class": "table2 condensed mt-5"})
        labels = table.find_all("td", limit=10)
        user_data = {"AC": int(labels[9].get_text()), "WA": int(labels[7].get_text()) - int(labels[9].get_text())}


        return title, user_data, submission_data

    def login(self):
        username = Config.get_kattis_name()
        token = Config.get_kattis_token()

        if(username == None or token == None):
            raise Exception("Kattis logging information not be found check your config file")
        
        login_url = "https://open.kattis.com/login"
        args = {
            'user': username, 
            'token': token,
            'script': 'true'
        }

        loader = Loader("Logging in...", "", color=Color("#00FFFFF"))
        loader.start()

        self.session = requests.Session()
        r = self.session.post(login_url, data=args)

        loader.stop()

        if(r.status_code != 200):
            raise Exception("Kattis logging information is not valid check your config file")
            
    def get_result(self):
        status = None
        loading_status = ["Compiling", "New", "Running"]

        loader = Loader("Getting result...", "", color=Color("#00FFFFF"))
        loader.start()
        while status == None or status in loading_status:
            time.sleep(0.75)
            r = self.session.get(self.result_url)
            parser = BeautifulSoup(r.text, "html.parser")
            status = parser.find("td", {"data-type": "status"}).get_text()
        loader.stop()

        time_result = parser.find("td", {"data-type": "cpu"}).get_text()
        cases = parser.find("td", {"data-type": "testcases"}).get_text()

        if status.find("Accepted") != -1:
            color = rgb(color_dic["AC"])
        else:
            color = rgb(color_dic["WA"])

        result_str = f"  {bold}{color}{status}{clear}"

        if(time_result != ""):
            result_str += f" {time_result}"
        
        if(cases != ""):
            result_str += f" Cases: {bold}{cases}{clear}"

        print_line()
        print(f"{bold}Result in problem {self.problem}:{clear}")
        print(result_str)

    def submit(self):
        self.login()

        data = {
            'submit': 'true',
            'submit_ctr': 2,
            'language': "C++",
            'mainclass': "",
            'problem': self.problem,
            'tag': "",
            'script': 'true'
        }

        files = []
        with open(Path(self.path, "main.cpp"), 'rb') as file:
            files.append((
                'sub_file[]', (
                    "main.cpp",
                    file.read(),
                    'application/octet-stream'
                )
            ))
        
        submit_url = "https://open.kattis.com/submit"
        loader = Loader("Submitting problem...", "", color=Color("#00FFFFF"))
        loader.start()
        r = self.session.post(submit_url, data=data, files=files)
        r.raise_for_status()
        loader.stop()

        if(r.text == "Problem not found"):
            raise CorruptedMetafileError
        
        self.result_url = r.text[5 + r.text.find("URL: "):]
        self.get_result()
